#include "global.h"
#include "NewNoteData.h"
#include "RageUtil.h"
#include "RageLog.h"


/*
 * This points to an entry in the NoteData mapping near row.  (row may 
 * not actually exist in the map.)  More precisely, this points to the
 * greatest row which is less than or equal to the above row at the time
 * this iterator was created.  For example, if rows 1, 5, and 10 exist,
 * and row=7, then this iterator points to 5.
 *
 * There are three cases: 1: a row exists that is <= m_iRow; 2: no such
 * row exists, but some rows > m_iRow do exist; or 3: there may be no rows
 * at all.  #1 is the normal case.  In #2, our iterator will point to a
 * value greater than m_iRow, so this case can be detected easily.  In
 * #3, our iterator will point to m_Data.end().
 *
 * (Design notes: in #3, we have no other choice.  In #2, we could point
 * at end() as well, but then a lot of cases would look like this:
 *   if( !m_Data.empty() && m_Iterator == m_Data.end() )
 *   {
 *       m_Iterator = m_Data.begin();
 *       ...
 *   }
 * We point at the next value, instead, so it simply looks like:
 *   if( m_Iterator != m_Data.end() && m_Iterator->first > m_Row )
 *   {
 *       ...
 *   }
 *
 * This can be searched for efficiently with UpdateRows().
 */

class ConstNoteDataIterator
{
public:
	ConstNoteDataIterator( const NewNoteData &n, int iRow = 0 );
	void Set( int iRow );

	/* Move to the next row that probably has data on it. */
	bool NextRow();
	bool PrevRow();
	bool operator++() { return NextRow(); }
	bool operator--() { return PrevRow(); }
	void AdvanceRows( int rows );

	/* Return true if we're pointing to a row that actually exists in the
	 * data (though it may not contain anything). */
	bool Exists() const { return m_Iterator != m_Data.end() && m_Iterator->first == m_Row; }
	bool AtEnd() const { return m_Iterator != m_Data.end(); }
	
	bool InHold( int iTrack ) const;
	ConstNoteDataIterator GetHoldEndRow( int iTrack ) const;
	int GetRow() const { return m_Row; }

	const NewNoteData::RowData &Get() const
	{
		ASSERT( Exists() );
		return m_Iterator->second;
	}
		
	

	// do in NoteDataUtil
	// bool RangeIsSparse( int RowFrom, int RowTo );

protected:
	void UpdateRows();

	const NewNoteData &m_NoteData;
	const NewNoteData::RowMap &m_Data;
	typedef NewNoteData::RowMap::const_iterator iter;
	iter m_Iterator;
	int m_Row;
};

class NoteDataIterator: public ConstNoteDataIterator
{
public:
	NoteDataIterator( NewNoteData &n, int iRow = 0 ):
		ConstNoteDataIterator( n, iRow ) { }

	NewNoteData::RowData &Get()
	{
		/* If it doesn't exist, create it.  Use the current iterator as a hint. */
		static const NewNoteData::RowData EmptyRow;
		NewNoteData::RowMap::value_type t( m_Row, EmptyRow );
		mutable_iter hint = GetIter();
		if( hint != m_Data.end() )
			++hint; /* see http://gcc.gnu.org/onlinedocs/libstdc++/23_containers/howto.html#4 */
		mutable_iter val = GetData().insert( hint, t ); 
		return val->second;
	}

private:
	typedef NewNoteData::RowMap::iterator mutable_iter;
	/* I'm assuming that all map iterators and const_iterators are really the
	 * same.  They are in every implementation I've seen. */
	mutable_iter GetIter() { return *(mutable_iter *)(&m_Iterator); }
	NewNoteData::RowMap &GetData() { return const_cast<NewNoteData::RowMap &>(m_Data); }
	
};


ConstNoteDataIterator::ConstNoteDataIterator( const NewNoteData &n, int iRow ):
	m_NoteData(n),
	m_Data(n.m_TapNotes)
{
	Set( iRow );
}

void ConstNoteDataIterator::Set( int iRow )
{
	m_Row = iRow;
	UpdateRows();
}

bool ConstNoteDataIterator::NextRow()
{
	if( m_Iterator == m_Data.end() )
		return false;

	++m_Iterator;
	m_Row = m_Iterator->first;
	return true;
}

/* Move to the previous row that probably has data on it. */
bool ConstNoteDataIterator::PrevRow()
{
	if( m_Iterator == m_Data.begin() )
		return false;

	--m_Iterator;
	m_Row = m_Iterator->first;
	return true;
}

/* Return true if the iterator lies within a hold for the given track. */
bool ConstNoteDataIterator::InHold( int iTrack ) const
{
	if( m_Iterator == m_Data.end() )
		return false;

	/* Iterate backwards until we find something on this track. */
	const NewNoteData::RowData &row = m_Iterator->second;
	if( row.m_Taps[iTrack] == TAP_HOLD_HEAD )
		return true;
	
	return false;
}

ConstNoteDataIterator ConstNoteDataIterator::GetHoldEndRow( int iTrack ) const
{
	if( m_Iterator == m_Data.end() )
		return m_Data.end();

	const NewNoteData::RowData &row = m_Iterator->second;
	if( row.m_Taps[iTrack] == TAP_HOLD_HEAD )
	{
		ConstNoteDataIterator ret = *this;
		while( !ret.AtEnd() )
		{
			++ret;
			const NewNoteData::RowData &row2 = m_Iterator->second;
			if( row2.m_Taps[iTrack] == TAP_HOLD_TAIL )
				break;
		}
		return ret;
	}
	if( Exists() && row.m_Taps[iTrack] == TAP_HOLD_TAIL )
		return true;
	
	return false;

}

void ConstNoteDataIterator::AdvanceRows( int iRows )
{
	m_Row += iRows;

	/* Data is typically sparse, so it's probably faster to iterate forward to
	 * find the next row than to do UpdateRows(). */ 
	if( iRows > 0 )
	{
		/* Iterate forward. */
		while( m_Iterator != m_Data.end() )
		{
			iter next = m_Iterator;
			++next;

			if( next->first > m_Row )
				break;

			m_Iterator = next;
		}
	}
	else if( iRows < 0 )
	{
		/* Iterate backward. */
		while( m_Iterator != m_Data.begin() && m_Iterator->first > m_Row )
			--m_Iterator;
	}
}

void ConstNoteDataIterator::UpdateRows()
{
	/* Find the first element greater than m_Row. */
	m_Iterator = m_Data.upper_bound( m_Row );
	
	if( m_Iterator == m_Data.begin() )
		return; /* There is no element less than m_Row. */

	/* There are; point to the last element. */
	--m_Iterator;
}

/* If the current row is within a hold note for the given track, cut the hold
 * note. */
void NewNoteData::CutHoldNote( NoteDataIterator &it, int track )
{
	if( !it.InHold(track) )
		return;

	it.AdvanceRows( -1 );
	{
		NewNoteData::RowData &row = it.Get();
		if( row.m_Taps[track] == TAP_HOLD_HEAD )
			row.m_Taps[track] = TAP_EMPTY;
		else
			row.m_Taps[track] = TAP_HOLD_TAIL;
	}

	it.AdvanceRows( +2 );
	{
		RowData &row = it.Get();
		if( row.m_Taps[track] == TAP_HOLD_TAIL )
			row.m_Taps[track] = TAP_EMPTY;
		else
			row.m_Taps[track] = TAP_HOLD_HEAD;
	}
	it.AdvanceRows( -1 );
}

NewNoteData::NewNoteData()
{
	m_iNumTracks = 0;
	Init();
}

void NewNoteData::Init()
{
	ClearAll();
	m_iNumTracks = 0;	// must do this after calling ClearAll()!
}

NewNoteData::~NewNoteData()
{
}

int NewNoteData::GetNumTracks() const
{
	return m_iNumTracks;
}

void NewNoteData::SetNumTracks( int iNewNumTracks )
{
	const int iOldNumTracks = m_iNumTracks;
	m_iNumTracks = iNewNumTracks;

	ASSERT( m_iNumTracks > 0 );

	// Make sure that all tracks are of the same length
	for( NoteDataIterator it( *this ); !it.AtEnd(); it.NextRow() )
	{
		RowData &row = it.Get();
		for( int t=iOldNumTracks; t<m_iNumTracks; ++t )
			row.m_Taps[t] = TAP_EMPTY;
	}
	
	
	/* Remove all hold notes that are out of bounds. */
	for( int h = m_HoldNotes.size()-1; h >= 0; --h )
		if( m_HoldNotes[h].iTrack >= iNewNumTracks )
			m_HoldNotes.erase( m_HoldNotes.begin()+h );
}


/* Clear [iNoteIndexBegin,iNoteIndexEnd]; that is, including iNoteIndexEnd. */
void NewNoteData::ClearRange( int iNoteIndexBegin, int iNoteIndexEnd )
{
	/* If iNoteIndexBegin-1 is within a hold note, it becomes the tail. */
	NoteDataIterator it( *this );
	it.Set( iNoteIndexBegin-1 );
	for( int t=0; t < m_iNumTracks; ++t )
		CutHoldNote( it, t );
	it.Set( iNoteIndexEnd+1 );
	for( int t=0; t < m_iNumTracks; ++t )
		CutHoldNote( it, t );
	
	RowMap::iterator start = m_TapNotes.lower_bound( iNoteIndexBegin );
	RowMap::iterator end = m_TapNotes.upper_bound( iNoteIndexEnd );
	
	m_TapNotes.erase( start, end );
}

void NewNoteData::ClearAll()
{
	m_TapNotes.clear();
	m_HoldNotes.clear();
}

/* Copy a range from pFrom to this.  (Note that this does *not* overlay;
 * all data in the range is overwritten.) */
void NewNoteData::CopyRange( const NewNoteData* pFrom, int iFromIndexBegin, int iFromIndexEnd, int iToIndexBegin )
{
	ASSERT( pFrom->m_iNumTracks == m_iNumTracks );

	const int iToIndexEnd = iToIndexBegin+iFromIndexEnd-iFromIndexBegin;
	this->ClearRange( iToIndexBegin, iToIndexEnd );
	
        ConstNoteDataIterator src( *pFrom, iFromIndexBegin );
        NoteDataIterator dst( *this, iToIndexBegin );
	for( ; !src.AtEnd() && src.GetRow() <= iToIndexEnd; ++src )
	{
		const int iToOffset = iToIndexBegin + (src.GetRow() - iFromIndexBegin);
		
		dst.Set( iToOffset );
		dst.Get() = src.Get();
	}
	
	/* Check the boundary for TAP_HOLD_HEAD that isn't followed by TAP_HOLD_TAIL.
	 * sequentialTAP_HOLD_HEAD,TAP_HOLD_HEAD and TAP_HOLD_TAIL,TAP_HOLD_TAIL,
	 * and . */
	dst.Set( iFromIndexBegin-1 );
	
	if( !dst.AtEnd() )
	{
		/* if row is HOLD_HEAD and row.next is not HOLD_TAIL, insert HOLD_TAIL at
		 * row+1. */
		RowData &row = dst.Get();
		for( int t=0; t < m_iNumTracks; ++t )
		{

			//XXX
		}
	}

	NewNoteData From, To;
	From.To4s( *pFrom );
	To.To4s( *this );

	// copy recorded TapNotes
	int f = iFromIndexBegin, t = iToIndexBegin;
	
	while( f<=iFromIndexEnd )
	{
		for( int c=0; c<m_iNumTracks; c++ )
		{
			TapNote tn = From.GetTapNote(c, f);
			if( IsTapAttack(tn) )
			{
				Attack attack = From.GetAttackAt(c,f);
				To.SetTapAttackNote( c, t, attack );
			}
			else
			{
				To.SetTapNote(c, t, tn);
			}
		}
		f++;
		t++;
	}

	this->From4s( To );
}

void NewNoteData::Config( const NewNoteData &From )
{
	m_iNumTracks = From.m_iNumTracks;
}

void NewNoteData::CopyAll( const NewNoteData* pFrom )
{
	Config(*pFrom);
	ClearAll();

	m_TapNotes = pFrom->m_TapNotes;
//	for( int c=0; c<m_iNumTracks; c++ )
//		m_TapNotes[c] = pFrom->m_TapNotes[c];
//	m_HoldNotes = pFrom->m_HoldNotes;
//	m_AttackMap = pFrom->m_AttackMap;
}

bool NewNoteData::IsRowEmpty( int index ) const
{
	/* If this is out of range, we don't have any notes there, so all tracks are empty. */
	if( index < 0 || index >= GetMaxRow() )
		return true;

	for( int t=0; t<m_iNumTracks; t++ )
		if( GetTapNoteX(t, index) != TAP_EMPTY )
			return false;
	return true;
}

bool NewNoteData::IsRangeEmpty( int track, int iIndexBegin, int iIndexEnd ) const
{
	ASSERT( track<m_iNumTracks );
	
	CLAMP( iIndexBegin, 0, GetMaxRow() );
	CLAMP( iIndexEnd, 0, GetMaxRow() );

	for( int i=iIndexBegin; i<=iIndexEnd; i++ )
		if( GetTapNoteX(track,i) != TAP_EMPTY )
			return false;
	return true;
}

int NewNoteData::GetNumTapNonEmptyTracks( int index ) const
{
	int iNum = 0;
	for( int t=0; t<m_iNumTracks; t++ )
		if( GetTapNote(t, index) != TAP_EMPTY )
			iNum++;
	return iNum;
}

void NewNoteData::GetTapNonEmptyTracks( int index, set<int>& addTo ) const
{
	for( int t=0; t<m_iNumTracks; t++ )
		if( GetTapNote(t, index) != TAP_EMPTY )
			addTo.insert(t);
}

int NewNoteData::GetFirstNonEmptyTrack( int index ) const
{
	/* If this is out of range, we don't have any notes there, so all tracks are empty. */
	if( index < 0 || index >= GetMaxRow() )
		return 0;

	for( int t=0; t<m_iNumTracks; t++ )
		if( GetTapNoteX( t, index ) != TAP_EMPTY )
			return t;
	return -1;
}

int NewNoteData::GetNumTracksWithTap( int index ) const
{
	/* If this is out of range, we don't have any notes there, so all tracks are empty. */
	if( index < 0 || index >= GetMaxRow() )
		return 0;

	int iNum = 0;
	for( int t=0; t<m_iNumTracks; t++ )
	{
		TapNote tn = GetTapNoteX( t, index );
		if( tn == TAP_TAP )
			iNum++;
	}
	return iNum;
}

int NewNoteData::GetNumTracksWithTapOrHoldHead( int index ) const
{
	/* If this is out of range, we don't have any notes there, so all tracks are empty. */
	if( index < 0 || index >= GetMaxRow() )
		return 0;

	int iNum = 0;
	for( int t=0; t<m_iNumTracks; t++ )
	{
		TapNote tn = GetTapNoteX( t, index );
		if( tn == TAP_TAP || tn == TAP_ADDITION || tn == TAP_HOLD_HEAD )
			iNum++;
	}
	return iNum;
}

int NewNoteData::GetFirstTrackWithTap( int index ) const
{
	/* If this is out of range, we don't have any notes there, so all tracks are empty. */
	if( index < 0 || index >= GetMaxRow() )
		return -1;

	for( int t=0; t<m_iNumTracks; t++ )
	{
		TapNote tn = GetTapNoteX( t, index );
		if( tn == TAP_TAP || tn == TAP_ADDITION )
			return t;
	}
	return -1;
}

int NewNoteData::GetFirstTrackWithTapOrHoldHead( int index ) const
{
	/* If this is out of range, we don't have any notes there, so all tracks are empty. */
	if( index < 0 || index >= GetMaxRow() )
		return -1;

	for( int t=0; t<m_iNumTracks; t++ )
	{
		TapNote tn = GetTapNoteX( t, index );
		if( tn == TAP_TAP || tn == TAP_ADDITION || tn == TAP_HOLD_HEAD )
			return t;
	}
	return -1;
}

void NewNoteData::AddHoldNote( HoldNote add )
{
	ASSERT( add.iStartRow>=0 && add.iEndRow>=0 );

	int i;

	// look for other hold notes that overlap and merge them
	// XXX: this is done implicitly with 4s, but 4s uses this function.
	// Rework this later.
	for( i=0; i<GetNumHoldNotes(); i++ )	// for each HoldNote
	{
		HoldNote &other = GetHoldNote(i);
		if( add.iTrack == other.iTrack  &&		// the tracks correspond
			add.RangeOverlaps(other) ) // they overlap
		{
			add.iStartRow = min(add.iStartRow, other.iStartRow);
			add.iEndRow = max(add.iEndRow, other.iEndRow);

			// delete this HoldNote
			RemoveHoldNote( i );
			--i;
		}
	}

	int iAddStartIndex = add.iStartRow;
	int iAddEndIndex = add.iEndRow;

	// delete TapNotes under this HoldNote
	for( i=iAddStartIndex+1; i<=iAddEndIndex; i++ )
		SetTapNote(add.iTrack, i, TAP_EMPTY);

	// add a tap note at the start of this hold
	SetTapNote(add.iTrack, iAddStartIndex, TAP_HOLD_HEAD);		// Hold begin marker.  Don't draw this, but do grade it.

	m_HoldNotes.push_back(add);
}

void NewNoteData::RemoveHoldNote( int iHoldIndex )
{
	ASSERT( iHoldIndex >= 0  &&  iHoldIndex < GetNumHoldNotes() );

	HoldNote& hn = GetHoldNote(iHoldIndex);

	const int iHoldStartIndex = hn.iStartRow;

	// delete a tap note at the start of this hold
	SetTapNote(hn.iTrack, iHoldStartIndex, TAP_EMPTY);

	// remove from list
	m_HoldNotes.erase(m_HoldNotes.begin()+iHoldIndex, m_HoldNotes.begin()+iHoldIndex+1);
}

/* Get a hold note with the same track and end row as hn. */
int NewNoteData::GetMatchingHoldNote( const HoldNote &hn ) const
{
	for( int i=0; i<GetNumHoldNotes(); i++ )	// for each HoldNote
	{
		const HoldNote &ret = GetHoldNote(i);
		if( ret.iTrack == hn.iTrack && ret.iEndRow == hn.iEndRow )
			return i;
	}
	FAIL_M( ssprintf("%i..%i, %i", hn.iStartRow, hn.iEndRow, hn.iTrack) );
}


void NewNoteData::SetTapAttackNote( int iTrack, int iRow, Attack attack )
{
//XXX	PruneUnusedAttacksFromMap();

	// find first unused attack index
/*	TapNote tn;
	for( tn = TAP_ATTACK_BEGIN; tn<=TAP_ATTACK_END; tn++ )
	{
		if( m_AttackMap.find(tn) == m_AttackMap.end() )	// this TapNote value is free to use
			goto done_searching;
	}
	ASSERT(0);

done_searching:
*/	

	RowData row = GetTapRow( iRow );
	row.m_Taps[iTrack] = TAP_ATTACK_BEGIN; /* XXX: rename to TAP_ATTACK */
	row.m_Attack[iTrack] = attack;
	SetTapRow( iRow, row );
}
#if 0
void NewNoteData::PruneUnusedAttacksFromMap()
{
	// Add all used AttackNote values to a map.
	map<TapNote,int> mapAttackToNothing;

	int max_row = GetMaxRow();
	for( int t=0; t<m_iNumTracks; t++ )
	{
		for( int r=0; r<=max_row; r++ )
		{
			TapNote tn = GetTapNote(t, r);
			if( IsTapAttack( tn ) )
				mapAttackToNothing[tn] = 1;
		}
	}

	// Remove all items from m_AttackMap that don't have corresponding
	// TapNotes in use.
	for( TapNote tn = TAP_ATTACK_BEGIN; tn<=TAP_ATTACK_END; tn++ )
	{
		bool bInAttackMap = m_AttackMap.find(tn) != m_AttackMap.end();
		bool bActuallyUsed = mapAttackToNothing.find(tn) != mapAttackToNothing.end();

		if( bActuallyUsed && !bInAttackMap )
			ASSERT( 0 );

		if( bInAttackMap && !bActuallyUsed )
		{
			m_AttackMap.erase( tn );
		}
	}
}
#endif
const Attack& NewNoteData::GetAttackAt( int iTrack, int iRow )
{
	RowData row = GetTapRow( iRow );
	row.m_Taps[iTrack] = TAP_ATTACK_BEGIN; /* XXX: rename to TAP_ATTACK */
	ASSERT( row.m_Taps[iTrack] == TAP_ATTACK_BEGIN );
	return row.m_Attack[iTrack];
}


int NewNoteData::GetFirstRow() const
{ 
	ConstNoteDataIterator it( *this );
	while( !it.AtEnd() && !it.Exists() )
		++it;

	if( it.AtEnd() ) // there are no notes
		return 0;

	return it.GetRow();
}

float NewNoteData::GetFirstBeat() const		
{ 
	return NoteRowToBeat( GetFirstRow() );
}

int NewNoteData::GetLastRow() const
{ 
	ConstNoteDataIterator it( *this, 999999 );
	--it;

	if( it.AtEnd() ) // there are no notes
		return 0;

	return it.GetRow();
}

float NewNoteData::GetLastBeat() const
{ 
	return NoteRowToBeat( GetLastRow() );
}

int NewNoteData::GetNumTapNotes( float fStartBeat, float fEndBeat ) const
{
	int iNumNotes = 0;

	if( fEndBeat == -1 )
		fEndBeat = GetMaxBeat();

	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );

	/* Clamp to known-good ranges. */
	iStartIndex = max( iStartIndex, 0 );
	iEndIndex = min( iEndIndex, GetMaxRow()-1 );
	
	for( int t=0; t<m_iNumTracks; t++ )
	{
		for( int i=iStartIndex; i<=iEndIndex; i++ )
		{
			TapNote tn = GetTapNoteX(t, i);
			if( tn != TAP_EMPTY && tn != TAP_MINE )
				iNumNotes++;
		}
	}
	
	return iNumNotes;
}

int NewNoteData::GetNumRowsWithTap( float fStartBeat, float fEndBeat ) const
{
	int iNumNotes = 0;

	if(fEndBeat == -1) fEndBeat = GetMaxBeat();
	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );
	
	for( int i=iStartIndex; i<=iEndIndex; i++ )
		if( IsThereATapAtRow(i) )
			iNumNotes++;
	
	return iNumNotes;
}

int NewNoteData::GetNumMines( float fStartBeat, float fEndBeat ) const
{
	int iNumMines = 0;

	if( fEndBeat == -1 )
		fEndBeat = GetMaxBeat();

	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );

	/* Clamp to known-good ranges. */
	iStartIndex = max( iStartIndex, 0 );
	iEndIndex = min( iEndIndex, GetMaxRow()-1 );
	
	for( int t=0; t<m_iNumTracks; t++ )
	{
		for( int i=iStartIndex; i<=iEndIndex; i++ )
			if( GetTapNoteX(t, i) == TAP_MINE )
				iNumMines++;
	}
	
	return iNumMines;
}

int NewNoteData::GetNumRowsWithTapOrHoldHead( float fStartBeat, float fEndBeat ) const
{
	int iNumNotes = 0;

	if(fEndBeat == -1) fEndBeat = GetMaxBeat();
	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );
	
	for( int i=iStartIndex; i<=iEndIndex; i++ )
		if( IsThereATapOrHoldHeadAtRow(i) )
			iNumNotes++;
	
	return iNumNotes;
}

int NewNoteData::RowNeedsHands( const int row ) const
{
	int iNumNotesThisIndex = 0;
	for( int t=0; t<m_iNumTracks; t++ )
	{
		TapNote tn = GetTapNoteX(t, row);
		if( tn == TAP_MINE || tn == TAP_EMPTY || tn == TAP_HOLD_TAIL ) // mines don't count
			continue;
		++iNumNotesThisIndex;
	}

	/* We must have at least one non-hold-body at this row to count it. */
	if( !iNumNotesThisIndex )
		return false;

	if( iNumNotesThisIndex < 3 )
	{
		/* We have at least one, but not enough.  Count holds. */
		for( int j=0; j<GetNumHoldNotes(); j++ )
		{
			const HoldNote &hn = GetHoldNote(j);
			if( hn.iStartRow+1 <= row && row <= hn.iEndRow )
				++iNumNotesThisIndex;
		}
	}

	return iNumNotesThisIndex >= 3;
}

int NewNoteData::GetNumHands( float fStartBeat, float fEndBeat ) const
{
	/* Count the number of times you have to use your hands.  This includes
	 * three taps at the same time, a tap while two hold notes are being held,
	 * etc.  Only count rows that have at least one tap note (hold heads count).
	 * Otherwise, every row of hold notes counts, so three simultaneous hold
	 * notes will count as hundreds of "hands". */
	if( fEndBeat == -1 )
		fEndBeat = GetMaxBeat();

	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );

	/* Clamp to known-good ranges. */
	iStartIndex = max( iStartIndex, 0 );
	iEndIndex = min( iEndIndex, GetMaxRow()-1 );

	int iNum = 0;
	for( int i=iStartIndex; i<=iEndIndex; i++ )
	{
		if( !RowNeedsHands(i) )
			continue;

		iNum++;
	}

	return iNum;
}

int NewNoteData::GetNumN( int MinTaps, float fStartBeat, float fEndBeat ) const
{
	if( fEndBeat == -1 )
		fEndBeat = GetMaxBeat();

	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );

	/* Clamp to known-good ranges. */
	iStartIndex = max( iStartIndex, 0 );
	iEndIndex = min( iEndIndex, GetMaxRow()-1 );

	int iNum = 0;
	for( int i=iStartIndex; i<=iEndIndex; i++ )
	{
		int iNumNotesThisIndex = 0;
		for( int t=0; t<m_iNumTracks; t++ )
		{
			TapNote tn = GetTapNoteX(t, i);
			if( tn != TAP_MINE && tn != TAP_EMPTY ) // mines don't count
				iNumNotesThisIndex++;
		}
		if( iNumNotesThisIndex >= MinTaps )
			iNum++;
	}
	
	return iNum;
}

int NewNoteData::GetNumHoldNotes( float fStartBeat, float fEndBeat ) const
{
	if( fEndBeat == -1 )
		fEndBeat = GetMaxBeat();
	int iStartIndex = BeatToNoteRow( fStartBeat );
	int iEndIndex = BeatToNoteRow( fEndBeat );

	int iNumHolds = 0;
	for( int i=0; i<GetNumHoldNotes(); i++ )
	{
		const HoldNote &hn = GetHoldNote(i);
		if( iStartIndex <= hn.iStartRow &&  hn.iEndRow <= iEndIndex )
			iNumHolds++;
	}
	return iNumHolds;
}

void NewNoteData::Convert2sAnd3sToHoldNotes()
{
	// Any note will end a hold (not just a TAP_HOLD_TAIL).  This makes parsing DWIs much easier.
	// Plus, allowing tap notes in the middle of a hold doesn't make sense!

	int rows = GetLastRow();
	for( int col=0; col<m_iNumTracks; col++ )	// foreach column
	{
		for( int i=0; i<=rows; i++ )	// foreach TapNote element
		{
			if( GetTapNote(col, i) != TAP_HOLD_HEAD )	// this is a HoldNote begin marker
				continue;
			SetTapNote(col, i, TAP_EMPTY);

			for( int j=i+1; j<=rows; j++ )	// search for end of HoldNote
			{
				// end hold on the next note we see
				if( GetTapNote(col, j) == TAP_EMPTY )
					continue;

				SetTapNote(col, j, TAP_EMPTY);

				AddHoldNote( HoldNote(col, i, j) );
				break;
			}
		}
	}
}


/* "102000301" ==
 * "104444001" */
void NewNoteData::ConvertHoldNotesTo2sAnd3s()
{
	// copy HoldNotes into the new structure, but expand them into 2s and 3s
	for( int i=0; i<GetNumHoldNotes(); i++ ) 
	{
		const HoldNote &hn = GetHoldNote(i);
		
		/* If they're the same, then they got clamped together, so just ignore it. */
		if( hn.iStartRow != hn.iEndRow )
		{
			SetTapNote( hn.iTrack, hn.iStartRow, TAP_HOLD_HEAD );
			SetTapNote( hn.iTrack, hn.iEndRow, TAP_HOLD_TAIL );
		}
	}
	m_HoldNotes.clear();
}


void NewNoteData::To2sAnd3s( const NewNoteData &out )
{
	CopyAll( &out );
	ConvertHoldNotesTo2sAnd3s();
}

void NewNoteData::From2sAnd3s( const NewNoteData &out )
{
	CopyAll( &out );
	Convert2sAnd3sToHoldNotes();
}

void NewNoteData::To4s( const NewNoteData &out )
{
	CopyAll( &out );
	ConvertHoldNotesTo4s();
}

void NewNoteData::From4s( const NewNoteData &out )
{
	CopyAll( &out );
	Convert4sToHoldNotes();
}

/* "104444001" ==
 * "102000301"
 *
 * "4441" basically means "hold for three rows then hold for another tap";
 * since taps don't really have a length, it's equivalent to "4440".
 * So, make sure the character after a 4 is always a 0. */
void NewNoteData::Convert4sToHoldNotes()
{
	int rows = GetLastRow();
	for( int col=0; col<m_iNumTracks; col++ )	// foreach column
	{
		for( int i=0; i<=rows; i++ )	// foreach TapNote element
		{
			if( GetTapNote(col, i) != TAP_HOLD )	// this is a HoldNote body
				continue;

			HoldNote hn( col, i, 0 );
			// search for end of HoldNote
			do {
				SetTapNote(col, i, TAP_EMPTY);
				i++;
			} while( GetTapNote(col, i) == TAP_HOLD);
			SetTapNote(col, i, TAP_EMPTY);

			hn.iEndRow = i;
			AddHoldNote( hn );
		}
	}
}

void NewNoteData::ConvertHoldNotesTo4s()
{
	// copy HoldNotes into the new structure, but expand them into 4s
	for( int i=0; i<GetNumHoldNotes(); i++ ) 
	{
		const HoldNote &hn = GetHoldNote(i);

		for( int j = hn.iStartRow; j < hn.iEndRow; ++j)
			SetTapNote(hn.iTrack, j, TAP_HOLD);
	}
	m_HoldNotes.clear();
}
/* Convert 
 *         vector<RowData> m_TapNotes[MAX_NOTE_TRACKS];
 *         */
void To4s( vector<RowData> &dst, const NewNoteData &src, int iStart, int iEnd )
{
	dst.clear();
	dst.insert( dst.begin(), iEnd-iStart+1, RowData() );

	for( ConstNoteDataIterator it( src, iStart ); !it.AtEnd() && it.GetRow() <= iEnd; ++it )
	{
		int iDestRow = it.GetRow() - iStart;
		const RowData &srcrow = src.Get();
		RowData &dstrow = dst[iDestRow];
		
		dstrow = srcrow;

		/* Expand holds to TAP_HOLD. */
		for( int t = 0; t < m_iNumTracks; ++t )
		{
			bool bInHold = false;
			int iEndRow = 0;
			if( it.GetRow() == iStart )
			{
				bInHold = it.InHold( t );
				ConstNoteDataIterator next_row( it );
				
				
			} else {
				bInHold = (srcrow.m_Taps[t] == TAP_HOLD_HEAD);
			}
			if( bInHold )
			{

			}
				
			/* If this is the first row, or if this is a hold head

		

	}


}

// -1 for iOriginalTracksToTakeFrom means no track
void NewNoteData::LoadTransformed( const NewNoteData* pOriginal, int iNewNumTracks, const int iOriginalTrackToTakeFrom[] )
{
	// reset all notes
	Init();
	
	Config( *pOriginal );
	m_iNumTracks = iNewNumTracks;

        ConstNoteDataIterator src( *pOriginal );
        NoteDataIterator dst( *this );

	for( ; !src.AtEnd(); ++src )
	{
		if( !src.Exists() )
			continue;

		const RowData &srcrow = dst.Get();
		RowData &dstrow = dst.Get();

		// copy tracks
		for( int t=0; t<m_iNumTracks; t++ )
		{
			const int iOriginalTrack = iOriginalTrackToTakeFrom[t];
			RAGE_ASSERT_M( iOriginalTrack < pOriginal->m_iNumTracks, ssprintf("from %i >= %i (to %i)", 
						iOriginalTrack, pOriginal->m_iNumTracks, iOriginalTrackToTakeFrom[t]));

			if( iOriginalTrack == -1 )
				continue;
			dstrow.CopyTrack( srcrow, iOriginalTrack, t );
		}
	}
}


void NewNoteData::LoadOverlapped( const NewNoteData* pOriginal, int iNewNumTracks )
{
	SetNumTracks( iNewNumTracks );

	NewNoteData in;
	in.To4s( *pOriginal );

	/* Keep track of the last source track that put a tap into each destination track,
	 * and the row of that tap.  Then, if two rows are trying to put taps into the
	 * same row within the shift threshold, shift the newcomer source row. */
	int LastSourceTrack[MAX_NOTE_TRACKS];
	int LastSourceRow[MAX_NOTE_TRACKS];
	int DestRow[MAX_NOTE_TRACKS];
	
	for( unsigned tr = 0; tr < MAX_NOTE_TRACKS; ++tr )
	{
		LastSourceTrack[tr] = -1;
		LastSourceRow[tr] = -9999;
		DestRow[tr] = tr;
		wrap( DestRow[tr], iNewNumTracks );
	}

	const int ShiftThreshold = BeatToNoteRow(1);

	const int iLastRow = in.GetMaxRow();
	for( int row = 0; row < iLastRow; ++row )
	{
		for ( int i = 0; i < in.GetNumTracks(); i++ )
		{
			const int iTrackFrom = i;
			int &iTrackTo = DestRow[i];

			const TapNote iStepFrom = in.m_TapNotes[iTrackFrom][row];
			if( iStepFrom == TAP_EMPTY )
				continue;

			if( LastSourceTrack[iTrackTo] != iTrackFrom &&
				row - LastSourceRow[iTrackTo] < ShiftThreshold )
			{
				/* This destination track is in use by a different source track.  Use the
				 * least-recently-used track. */
				for( int DestTrack = 0; DestTrack < iNewNumTracks; ++DestTrack )
					if( LastSourceRow[DestTrack] < LastSourceRow[iTrackTo] )
						iTrackTo = DestTrack;
			}

			/* If it's still in use, then we just don't have an available track. */
			if( LastSourceTrack[iTrackTo] != iTrackFrom &&
				row - LastSourceRow[iTrackTo] < ShiftThreshold )
				continue;

			LastSourceTrack[iTrackTo] = iTrackFrom;
			LastSourceRow[iTrackTo] = row;
			SetTapNote( iTrackTo, row, iStepFrom );
		}
	}

	Convert4sToHoldNotes();
}

void NewNoteData::LoadTransformedSlidingWindow( const NewNoteData* pOriginal, int iNewNumTracks )
{
	// reset all notes
	Init();

	if( pOriginal->GetNumTracks() > iNewNumTracks )
	{
		/* Use a different algorithm for reducing tracks. */
		LoadOverlapped( pOriginal, iNewNumTracks );
		return;
	}

	NewNoteData Original;
	Original.To4s( *pOriginal );

	Config(*pOriginal);
	m_iNumTracks = iNewNumTracks;


	int iCurTrackOffset = 0;
	int iTrackOffsetMin = 0;
	int iTrackOffsetMax = abs( iNewNumTracks - Original.m_iNumTracks );
	int bOffsetIncreasing = true;

	int iLastRow = Original.GetLastRow();
	for( int r=0; r<=iLastRow; )
	{
		// copy notes in this measure
		for( int t=0; t<Original.m_iNumTracks; t++ )
		{
			int iOldTrack = t;
			int iNewTrack = (iOldTrack + iCurTrackOffset) % iNewNumTracks;
			this->SetTapNote(iNewTrack, r, Original.GetTapNote(iOldTrack, r));
		}
		r++;

		if( r % (ROWS_PER_MEASURE*4) == 0 )	// adjust sliding window every 4 measures
		{
			// See if there is a hold crossing the beginning of this measure
			bool bHoldCrossesThisMeasure = false;

			if( r )
			for( int t=0; t<=Original.m_iNumTracks; t++ )
			{
				if( Original.GetTapNote(t, r) == TAP_HOLD &&
					Original.GetTapNote(t, r-1) == TAP_HOLD)
				{
					bHoldCrossesThisMeasure = true;
					break;
				}
			}

			// adjust offset
			if( !bHoldCrossesThisMeasure )
			{
				iCurTrackOffset += bOffsetIncreasing ? 1 : -1;
				if( iCurTrackOffset == iTrackOffsetMin  ||  iCurTrackOffset == iTrackOffsetMax )
					bOffsetIncreasing ^= true;
				CLAMP( iCurTrackOffset, iTrackOffsetMin, iTrackOffsetMax );
			}
		}
	}

	Convert4sToHoldNotes();

	m_AttackMap = Original.GetAttackMap();
}

void NewNoteData::PadTapNotes(int rows)
{
	int needed = rows - m_TapNotes[0].size() + 1;
	if(needed < 0) 
		return;

	needed += 100; /* optimization: give it a little more than it needs */

	for(int track = 0; track < m_iNumTracks; ++track)
		m_TapNotes[track].insert(m_TapNotes[track].end(), needed, TAP_EMPTY);
}

void NewNoteData::MoveTapNoteTrack(int dest, int src)
{
	if(dest == src) return;
	m_TapNotes[dest] = m_TapNotes[src];
	m_TapNotes[src].clear();
}

void NewNoteData::SetTapNote(int track, int row, TapNote t)
{
	if(row < 0) return;

	PadTapNotes(row);
	m_TapNotes[track][row]=t;
}

void NewNoteData::ReserveRows( int row )
{
	for(int track = 0; track < m_iNumTracks; ++track)
		m_TapNotes[track].reserve( row );
}

void NewNoteData::EliminateAllButOneTap(int row)
{
	if(row < 0) return;

	PadTapNotes(row);

	int track;
	for(track = 0; track < m_iNumTracks; ++track)
	{
		if( m_TapNotes[track][row] == TAP_TAP )
			break;
	}

	track++;

	for( ; track < m_iNumTracks; ++track)
	{
		if( m_TapNotes[track][row] == TAP_TAP )
			m_TapNotes[track][row] = TAP_EMPTY;
	}
}

void NewNoteData::GetTracksHeldAtRow( int row, set<int>& addTo )
{
	for( unsigned i=0; i<m_HoldNotes.size(); i++ )
	{
		const HoldNote& hn = m_HoldNotes[i];
		if( hn.RowIsInRange(row) )
			addTo.insert( hn.iTrack );
	}
}

int NewNoteData::GetNumTracksHeldAtRow( int row )
{
	static set<int> viTracks;
	viTracks.clear();
	GetTracksHeldAtRow( row, viTracks );
	return viTracks.size();
}

/*
 * (c) 2001-2004 Chris Danford, Glenn Maynard
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, provided that the above
 * copyright notice(s) and this permission notice appear in all copies of
 * the Software and that both the above copyright notice(s) and this
 * permission notice appear in supporting documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF
 * THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS
 * INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

